@@ -17,7 +17,9 @@ import os
 from pathlib import Path, PurePosixPath
 import posixpath
 import shutil
+import stat
 import threading
+import time
 from types import TracebackType
 from typing import (
     IO,
@@ -26,10 +28,12 @@ from typing import (
     Callable,
     Dict,
     Iterable,
+    Iterator,
     List,
     Optional,
     Sequence,
     Set,
+    Tuple,
     Type,
     Union,
     cast,
@@ -43,6 +47,7 @@ __all__ = (
     "ZipFileExtra",
     "FilteredZipInfo",
     "StopZipIndexRead",
+    "open_file_in_zip",
     "read_file_in_zip",
     "extract_file_in_zip",
     "NOTSET",
@@ -216,7 +221,7 @@ class ZipPath:
     # pathlib like interface
 
     @property
-    def name(self):
+    def name(self) -> str:
         """Return the basename of the current internal path within the zip file."""
         return posixpath.basename(self.at)
 
@@ -226,6 +231,11 @@ class ZipPath:
         parent_at = posixpath.dirname(self.at)
         return self.__class__(self, at=parent_at)
 
+    @property
+    def parts(self) -> Tuple[str, ...]:
+        """Return the parts of the current internal path within the zip file."""
+        return PurePosixPath(self.at).parts
+
     def is_dir(self):
         """Whether this path is an existing directory."""
         return self.exists() and not self.is_file()
@@ -279,7 +289,7 @@ class ZipPath:
         # zip file open misleading signals 'r', 'w', when actually they are byte mode
         zinfo: Union[str, zipfile.ZipInfo]
         if mode == "wb":
-            zinfo = zipfile.ZipInfo(self.at)
+            zinfo = zipfile.ZipInfo(self.at, date_time=time.localtime(time.time())[0:6])
             zinfo.compress_type = (
                 self._zipfile.compression if compression is NOTSET else compression
             )
@@ -325,6 +335,39 @@ class ZipPath:
         content = self.read_bytes()
         return content.decode(encoding=encoding)
 
+    def mkdir(self, mode=0o777, exist_ok=False):
+        """Create a new directory at this given path."""
+        if self.exists():
+            if not exist_ok:
+                raise FileExistsError(f"cannot create directory: '{self.at}'")
+            return
+        stat.filemode(mode)  # check the mode is valid
+        path = self.at if self.at.endswith("/") else self.at + "/"
+        # this code is copied directly from zipfile.ZipFile.write
+        zinfo = zipfile.ZipInfo(path, date_time=time.localtime(time.time())[0:6])
+        zinfo.file_size = 0
+        # TODO the mode is not currently used with ZipFile.extract
+        zinfo.external_attr = (mode & 0xFFFF) << 16  # Unix attributes
+        zinfo.external_attr |= 0x10  # MS-DOS directory flag
+        zinfo.compress_size = 0
+        zinfo.CRC = 0
+        _zipfile = self._zipfile
+        with _zipfile._lock:
+            if _zipfile._seekable:
+                _zipfile.fp.seek(_zipfile.start_dir)
+            zinfo.header_offset = _zipfile.fp.tell()  # Start of header bytes
+            if zinfo.compress_type == zipfile.ZIP_LZMA:
+                # Compressed data includes an end-of-stream (EOS) marker
+                zinfo.flag_bits |= 1 << 1
+
+            _zipfile._writecheck(zinfo)  # type: ignore[attr-defined]
+            _zipfile._didModify = True
+
+            _zipfile.filelist.append(zinfo)
+            _zipfile.NameToInfo[zinfo.filename] = zinfo
+            _zipfile.fp.write(zinfo.FileHeader(False))
+            _zipfile.start_dir = _zipfile.fp.tell()
+
     def iterdir(self) -> Iterable["ZipPath"]:
         """Iterate over the files and folders in this directory (non-recursive)."""
         if self.is_file():
@@ -550,13 +593,15 @@ class FilteredZipInfo(abc.MutableMapping):
         self._dict: Dict[str, zipfile.ZipInfo] = {}
         self._filenames = set(filenames)
         self._max_infos = max_infos
-        self._read = 0
+        self._read = 0.0
 
     def __getitem__(self, name):
         return self._dict.__getitem__(name)
 
     def __setitem__(self, name, item):
-        self._read += 1
+        self._read += (
+            1 / 2
+        )  # _RealGetContents appends to file list, then adds to mapping
         if name in self._filenames:
             self._dict.__setitem__(name, item)
         if self._max_infos and self._max_infos <= self._read:
@@ -734,6 +779,45 @@ class ZipFileExtra(zipfile.ZipFile):
         return list(self.NameToInfo)
 
 
+@contextmanager
+def open_file_in_zip(
+    filepath: str,
+    path: str,
+    *,
+    search_limit: Optional[int] = None,
+) -> Iterator[IO[bytes]]:
+    """Open a file from inside a zip file.
+
+    This function is optimised for cpu/memory performance,
+    since it returns the file as soon as its index is found in the zip registry,
+    and does not construct the full index in memory.
+
+    For best performance, the path should have been stored near the start of the index.
+
+    :param filepath: the path to the zip file
+    :param path: the relative path within the zip file
+    :param search_limit: Limit the search in the zip to the first n records
+
+    :raises IOError: If the zip file cannot be read
+    :raises FileNotFoundError: If the path in the zip file does not exist
+    """
+    try:
+        with ZipFileExtra(
+            filepath,
+            "r",
+            allowZip64=True,
+            name_to_info=FilteredZipInfo({path}, max_infos=search_limit),
+        ).open(path, "r") as handle:
+            yield handle
+    except zipfile.BadZipfile as error:
+        raise IOError(f"The input file cannot be read: {error}")
+    except KeyError:
+        msg = f"required file {path} is not included"
+        if search_limit:
+            msg += f" within first {search_limit} records"
+        raise FileNotFoundError(msg)
+
+
 def read_file_in_zip(
     filepath: str,
     path: str,
@@ -741,7 +825,7 @@ def read_file_in_zip(
     *,
     search_limit: Optional[int] = None,
 ) -> Union[bytes, str]:
-    """Read a  file from inside a zip file and return its content.
+    """Read a file from inside a zip file and return its content.
 
     This function is optimised for cpu/memory performance,
     since it returns the file as soon as its index is found in the zip registry,
@@ -758,20 +842,11 @@ def read_file_in_zip(
     :raises FileNotFoundError: If the path in the zip file does not exist
 
     """
-    try:
-        output = ZipFileExtra(
-            filepath,
-            "r",
-            allowZip64=True,
-            name_to_info=FilteredZipInfo({path}, max_infos=search_limit),
-        ).read(path)
-        if encoding is not None:
-            return output.decode(encoding)
-        return output
-    except zipfile.BadZipfile as error:
-        raise IOError(f"The input file cannot be read: {error}")
-    except KeyError:
-        raise FileNotFoundError(f"required file {path} is not included")
+    with open_file_in_zip(filepath, path, search_limit=search_limit) as zip_handle:
+        output = zip_handle.read()
+    if encoding is not None:
+        return output.decode(encoding)
+    return output
 
 
 def extract_file_in_zip(
@@ -800,17 +875,7 @@ def extract_file_in_zip(
     :raises FileNotFoundError: If the path in the zip file does not exist
 
     """
-    try:
-        with ZipFileExtra(
-            filepath,
-            "r",
-            allowZip64=True,
-            name_to_info=FilteredZipInfo({path}, max_infos=search_limit),
-        ).open(path) as zip_handle:
-            shutil.copyfileobj(
-                zip_handle, handle, **({"length": buffer_size} if buffer_size else {})
-            )
-    except zipfile.BadZipfile as error:
-        raise IOError(f"The zip file cannot be read: {error}")
-    except KeyError:
-        raise FileNotFoundError(f"required file {path} is not included")
+    with open_file_in_zip(filepath, path, search_limit=search_limit) as zip_handle:
+        shutil.copyfileobj(
+            zip_handle, handle, **({"length": buffer_size} if buffer_size else {})
+        )
